/*		Character grid hypertext object
**		===============================
*/


/* lynx specific defines */
#include "defs.h"

#include <assert.h>
#include <ctype.h>
#include "HTUtils.h"
#include "HTString.h"
#include "GridText.h"
#include "HTFont.h"
#include "HTBrowse.h"

struct _HTStream {                      /* only know it as object */
    CONST HTStreamClass *       isa;
    /* ... */
};

#define MAX_LINE	180	/* Hope that no widow is larger than this */
#ifdef THINK_C
#define LOADED_LIMIT 3			/* For now, save last two texts */
#else
#define LOADED_LIMIT 11			/* For now, save last ten texts */
#endif

#define TITLE_LINES  1


/*	From default style sheet:
*/
extern HTStyleSheet * styleSheet;	/* Default or overridden */


/*	Exports
*/ 
PUBLIC HText * HTMainText = 0;		/* Equivalent of main window */
PUBLIC HTParentAnchor * HTMainAnchor = 0;	/* Anchor for HTMainText */

PUBLIC char * HTAppName = "Lynx";      /* Application name */
PUBLIC char * HTAppVersion = "2.0.10";        /* Application version */

typedef struct _line {
	struct _line	*next;
	struct _line	*prev;
	short unsigned	offset;		/* Implicit initial spaces */
	short unsigned	size;		/* Number of characters */
	BOOL	split_after;		/* Can we split after? */
	BOOL	bullet;			/* Do we bullet? */
	char	data[1];		/* Space for terminator at least! */
} HTLine;

#define LINE_SIZE(l) (sizeof(HTLine)+(l))	/* allow for terminator */

typedef struct _TextAnchor {
	struct _TextAnchor *	next;
	int			number;		/* For user interface */
	int			start;		/* Characters */
        int			line_pos;       /* position in text */
	int			extent;		/* Characters */
	int			line_num;       /* place in document */
	char *		        hightext;       /* the link text */
	BOOL			show_anchor;    /* show the anchor? */
	BOOL			numbered;       /* is the anchor numbered */
	HTChildAnchor *		anchor;
} TextAnchor;


/*	Notes on struct _Htext:
**	next_line is valid iff state is false.
**	top_of_screen line means the line at the top of the screen
**			or just under the title if there is one.
*/
struct _HText {
	HTParentAnchor *	node_anchor;
	char *			title;
	HTLine * 		last_line;
	int			lines;		/* Number of them */
	int			chars;		/* Number of them */
	TextAnchor *		first_anchor;	/* Singly linked list */
	TextAnchor *		last_anchor;
	int			last_anchor_number;	/* user number */
/* For Internal use: */	
	HTStyle *		style;			/* Current style */
	int			display_on_the_fly;	/* Lines left */
	int			top_of_screen;		/* Line number */
	HTLine *		top_of_screen_line;	/* Top */
	HTLine *		next_line;		/* Bottom + 1 */
	int			permissible_split;	/* in last line */
	BOOL			in_line_1;		/* of paragraph */
	BOOL			stale;			/* Must refresh */

        HTStream*               target;                 /* Output stream */
        HTStreamClass           targetClass;            /* Output routines */
};

/*	Boring static variable used for moving cursor across
*/

#define SPACES(n) (&space_string[(MAX_LINE-1) - (n)])
                                   /* String containing blank spaces only */
PRIVATE char * space_string;

PRIVATE HTStyle default_style =
	{ 0,  "(Unstyled)", "",
	(HTFont)0, 1.0, HT_BLACK,		0, 0,
	0, 0, 0, HT_LEFT,		1, 0,	0, 
	NO, NO, 0, 0,			0 };	



PRIVATE HTList * loaded_texts;	/* A list of all those in memory */

/*			Creation Method
**			---------------
*/
PUBLIC HText *	HText_new ARGS1(HTParentAnchor *,anchor)
{
    HTLine * line;
    HText * self = (HText *) malloc(sizeof(*self));
    if (!self) return self;
    
    if (!loaded_texts) loaded_texts = HTList_new();
    HTList_addObject(loaded_texts, self);
    if (HTList_count(loaded_texts) >= LOADED_LIMIT) {
        if (TRACE) fprintf(stderr, "GridText: Freeing off cached doc.\n"); 
        HText_free((HText *)HTList_removeFirstObject(loaded_texts));
    }
    
    line = self->last_line = (HTLine *)malloc(LINE_SIZE(MAX_LINE));
    if (line == NULL) outofmem(__FILE__, "HText_New");
    line->next = line->prev = line;
    line->offset = line->size = 0;
    self->lines = self->chars = 0;
    self->title = 0;
    self->first_anchor = self->last_anchor = 0;
    self->style = &default_style;
    self->top_of_screen = 0;
    self->node_anchor = anchor;
    self->last_anchor_number = 0;	/* Numbering of them for references */
    self->stale = YES;
    
    HTAnchor_setDocument(anchor, (HyperDoc *)self);

    clear();
    HTMainText = self;
    HTMainAnchor = anchor;
    self->display_on_the_fly = 0;
    
    if (!space_string) {	/* Make a blank line */
        char *p;
	space_string = (char *)malloc(MAX_LINE+1);
	if (space_string == NULL) outofmem(__FILE__, "HText_New");
        for (p=space_string; p<space_string+(MAX_LINE-1); p++) 
            *p = ' '; 		/* Used for printfs later */
        space_string[(MAX_LINE-1)] = '\0'; 
    }
    
    return self;
}

/*                      Creation Method 2
**                      ---------------
**
**      Stream is assumed open and left open.
*/
PUBLIC HText *  HText_new2 ARGS2(
                HTParentAnchor *,       anchor,
                HTStream*,              stream)

{
    HText * this = HText_new(anchor);

    if (stream) {
        this->target = stream;
        this->targetClass = *stream->isa;       /* copy action procedures */
    }
    return this;
}

/*	Free Entire Text
**	----------------
*/
PUBLIC void 	HText_free ARGS1(HText *,self)
{
    HTAnchor_setDocument(self->node_anchor, (HyperDoc *)0);
    
    while(YES) {		/* Free off line array */
        HTLine * l = self->last_line;
	l->next->prev = l->prev;
	l->prev->next = l->next;	/* Unlink l */
	self->last_line = l->prev;
	free(l);
	if (l == self->last_line) break;	/* empty */
    };
    
    while(self->first_anchor) {		/* Free off anchor array */
        TextAnchor * l = self->first_anchor;
	self->first_anchor = l->next;
	free(l);
    }
    free(self);
}

/*		Display Methods
**		---------------
*/


/*	Output a line
**	-------------
*/
PRIVATE void display_line ARGS1(HTLine *,line)
{
      char argument[20];
      char *cp;
      /* make sure that we don't go over the COLS limit on the display! */

      /* make a printw argument of the form %s%.#s where 
       * # equals the number of cols left on the screen
       * after the offset is subtracted
       */
      sprintf(argument,"%%s%%.%ds\n",(COLS-1)-line->offset);
      printw(argument, SPACES(line->offset), line->data);

      /*  old stuff
       * addstr(SPACES(line->offset));
       * addstr(line->data);
       * addch('\n');
       */
   
}

/*	Output the title line
**	---------------------
*/
PRIVATE void display_title ARGS1(HText *,text)
{
    CONST char * title = HTAnchor_title(text->node_anchor);
    char percent[20], format[20];

    if (text->lines > (LINES-2)) {
	sprintf(percent, " (p%d of %d)",
	    (text->top_of_screen/(LINES-2)) + 1,
	    (text->lines-1)/(LINES-2) + 1);
    } else {
	percent[0] = 0;	/* Null string */
    }

    sprintf(format, "%%%d.%ds%%s\n",	/* Generate format string */
		    (COLS-1)-strlen(percent),
		    (COLS-1)-strlen(percent));
    if (TRACE) fprintf(stderr, "FORMAT IS `%s'\n", format);

     move(0,0);
     printw(format, title ? title : "" , percent);
}



/*	Output a page
**	-------------
*/
PRIVATE void display_page ARGS3(HText *,text, int,line_number, char *, target)
{
    HTLine * line = text->last_line->prev;
    int i;
    char *cp;
    CONST char * title = HTAnchor_title(text->node_anchor);
    int lines_of_text = title ? (LINES-TITLE_LINES)-1 : LINES-1;
    int last_screen = text->lines - lines_of_text;
    TextAnchor *Anchor_ptr;
    BOOL display_flag=FALSE;

    if(TRACE)
	fprintf(stderr,"GridText: Entering display_page\n");

/*	Constrain the line number to be within the document
*/
    if (text->lines <= lines_of_text) line_number = 0;
    else if (line_number>text->lines-1) line_number = last_screen;
    else if (line_number < 0) line_number = 0;
    
    for(i=0,  line = text->last_line->next;		/* Find line */
    	i<line_number && (line!=text->last_line);
      i++, line=line->next) /* Loop */ assert(line->next != NULL);

    /* clear the screen */
    clear();						/* No */
    text->top_of_screen = line_number;
    display_title(text);
    display_flag=TRUE;
    
 /*	print it
 */
    if (line) {
      for(i=0;
	  (i< lines_of_text) && (line != text->last_line); i++)  {
        assert(line != NULL);
        display_line(line);

        /* if the target is on this line, bold it */
        if(strlen(target) > 0 &&
		(case_sensitive ?  
		    (cp = strstr(line->data, target)) != NULL : 
		    (cp = mystrstr(line->data, target)) != NULL) &&
                ((cp-line->data) + line->offset + strlen(target)) < COLS) {
            move(i+1, line->offset + (cp-line->data));
	    start_underline();
	    addstr(target);
	    stop_underline();
	    move(i+2, 0);
	}

	display_flag=TRUE;
	line = line->next;
      }
    }

    text->next_line = line;	/* Line after screen */
    text->stale = NO;		/* Display is up-to-date */

    /* turn on bolding of links */
    start_bold();

    /* add the anchors to lynx structures */
    nlinks = 0;
    for(Anchor_ptr=text->first_anchor;  Anchor_ptr != NULL; 
					    Anchor_ptr = Anchor_ptr->next) {

	if(TRACE)
	    fprintf(stderr,"Anchor on line %d\n",Anchor_ptr->line_num);

	if(Anchor_ptr->line_num >= line_number &&
		Anchor_ptr->line_num < line_number+lines_of_text) {

	    if(Anchor_ptr->show_anchor) {

		 /* if the anchor is numbered, don't use the number
		  * as part of the highlighted text
		  */
		if(Anchor_ptr->numbered) {
			/* skip the number '[#]' */
     	       	    cp = strchr(Anchor_ptr->hightext, ']') + 1; 
		    strcpy(links[nlinks].hightext, cp);
		} else {
		    strcpy(links[nlinks].hightext, Anchor_ptr->hightext);
		}

		*links[nlinks].taddr = '\0';  /* set telnet address to null */

		links[nlinks].www_destination = HTAnchor_followMainLink(
					     (HTAnchor *)Anchor_ptr->anchor);
     	        strcpy(links[nlinks].lname, 
			HTAnchor_address(links[nlinks].www_destination) );


      	        links[nlinks].lx= Anchor_ptr->line_pos;
		/* if anchors are numbered we must compensate for the 
		 * x displacement caused by the number.
		 * cp is a pointer to the first charactor after
		 * the number that we want to skip
		 */
		if(Anchor_ptr->numbered) 
		    links[nlinks].lx += cp - Anchor_ptr->hightext;

      	        links[nlinks].ly= (Anchor_ptr->line_num+1)-line_number;
	        links[nlinks].type = WWW_LINK_TYPE;
		*links[nlinks].target = '\0';

	        /* bold the link */
	        move(links[nlinks].ly, links[nlinks].lx);
	        addstr(links[nlinks].hightext);
   
		/* print some diagnostics */
		if(TRACE) {
	    	    fprintf(stderr,"GridText: hightext = %s\n",
						links[nlinks].hightext);
	    	    fprintf(stderr,"GridText: Anchor address = %s\n",
			     			links[nlinks].lname);
	 	}

	        nlinks++;

		display_flag = TRUE;

	    } else { /* not showing anchor */
		if(TRACE) 
		    fprintf(stderr,"GridText: Not showing link\n");
	    }
	} 

	if(Anchor_ptr == text->last_anchor || 
			Anchor_ptr->line_num > line_number+lines_of_text)
	    break;
    }

	    
    /* turn off bolding */	
    stop_bold();

    if(!display_flag) /* nothing on the page */
	addstr("\n     Document is empty");

    refresh();

}


/*			Object Building methods
**			-----------------------
**
**	These are used by a parser to build the text in an object
*/
PUBLIC void HText_beginAppend ARGS1(HText *,text)
{
    text->permissible_split = 0;
    text->in_line_1 = YES;
}


/*	Add a new line of text
**	----------------------
**
** On entry,
**
**	split	is zero for newline function, else number of characters
**		before split.
**	text->display_on_the_fly
**		may be set to indicate direct output of the finished line.
** On exit,
**		A new line has been made, justified according to the
**		current style. Text after the split (if split nonzero)
**		is taken over onto the next line.
**
**		If display_on_the_fly is set, then it is decremented and
**		the finished line is displayed.
*/
#define new_line(text) split_line(text, 0)

PRIVATE void split_line ARGS2(HText *,text, int,split)
{
    HTStyle * style = text->style;
    int spare;
    int indent = text->in_line_1 ? text->style->indent1st
    				 : text->style->leftIndent;
    
/*	Make new line
*/
    HTLine * previous = text->last_line;
    HTLine * line = (HTLine *) malloc(LINE_SIZE(MAX_LINE));
    
    if (line == NULL) outofmem(__FILE__, "split_line");
    text->lines++;
    
    previous->next->prev = line;
    line->prev = previous;
    line->next = previous->next;
    previous->next = line;
    text->last_line = line;
    line->size = 0;
    line->offset = 0;

/*	Split at required point
*/    
    if (split) {	/* Delete space at "split" splitting line */
	char * p;
	previous->data[previous->size] = 0;
	for (p = &previous->data[split]; *p; p++)
	    if (*p != ' ') break;
	strcpy(line->data, p);
	line->size = strlen(line->data);
	previous->size = split;
    }
    
/*	Economise on space.
**	Not on the RS6000 due to a chaotic bug in realloc argument passing.
**	Same problem with Ultrix (4.2) : realloc() is not declared properly.
*/
#ifndef AIX
#ifndef ultrix
    while ((previous->size > 0) &&
    	(previous->data[previous->size-1] == ' '))	/* Strip trailers */
        previous->size--;
	
    previous = (HTLine *) realloc (previous, LINE_SIZE(previous->size));
    if (previous == NULL) outofmem(__FILE__, "split_line");
#endif /* ultrix */
#endif /* AIX */

    previous->prev->next = previous;	/* Link in new line */
    previous->next->prev = previous;	/* Could be same node of course */

/*	Terminate finished line for printing
*/
    previous->data[previous->size] = 0;
     
    
/*	Align left, right or center
*/

    spare =  (COLS-1) -
    		style->rightIndent + style->leftIndent -
    		previous->size;	/* @@ first line indent */
		
    switch (style->alignment) {
	case HT_CENTER :
	    previous->offset = previous->offset + indent + spare/2;
	    break;
	case HT_RIGHT :
	    previous->offset = previous->offset + indent + spare;
	    break;
	case HT_LEFT :
	case HT_JUSTIFY :		/* Not implemented */
	default:
	    previous->offset = previous->offset + indent;
	    break;
    } /* switch */

    text->chars = text->chars + previous->size + 1;	/* 1 for the line */
    text->in_line_1 = NO;		/* unless caller sets it otherwise */
    
#ifdef DISPLAY_ON_FLY
/*	If displaying as we go, output it:
*/
    if (text->display_on_the_fly) {
        if ((text->display_on_the_fly == LINES-1) &&
		    HTAnchor_title(text->node_anchor)) { /* First line */
	    display_title(text);
	    text->display_on_the_fly--;
	}
	display_line(previous);
	text->display_on_the_fly--;
    }
#endif /* DISPLAY_ON_FLY */
} /* split_line */


/*	Allow vertical blank space
**	--------------------------
*/
PRIVATE void blank_lines ARGS2(HText *,text, int,newlines)
{
    if (text->last_line->size == 0) {	/* No text on current line */
	HTLine * line = text->last_line->prev;
	while ((line!=text->last_line) && (line->size == 0)) {
	    if (newlines==0) break;
	    newlines--;		/* Don't bother: already blank */
	    line = line->prev;
	}
    } else {
	newlines++;			/* Need also to finish this line */
    }

    for(;newlines;newlines--) {
	new_line(text);
    }
    text->in_line_1 = YES;
}


/*	New paragraph in current style
**	------------------------------
** See also: setStyle.
*/

PUBLIC void HText_appendParagraph ARGS1(HText *,text)
{
    int after = text->style->spaceAfter;
    int before = text->style->spaceBefore;
    blank_lines(text, after>before ? after : before);
}


/*	Set Style
**	---------
**
**	Does not filter unnecessary style changes.
*/
PUBLIC void HText_setStyle ARGS2(HText *,text, HTStyle *,style)
{
    int after, before;

    if (!style) return;				/* Safety */
    after = text->style->spaceAfter;
    before = style->spaceBefore;
    if (TRACE) fprintf(stderr, "HTML: Change to style %s\n", style->name);

    blank_lines (text, after>before ? after : before);

    text->style = style;
}


/*	Append a character to the text object
**	-------------------------------------
*/
PUBLIC void HText_appendCharacter ARGS2(HText *,text, char,ch)
{
    HTLine * line = text->last_line;
    HTStyle * style = text->style;
    int indent = text->in_line_1 ? style->indent1st : style->leftIndent;
    
/*		New Line
*/
    if (ch == '\n') {
	    new_line(text);
	    text->in_line_1 = YES;	/* First line of new paragraph */
	    return;
    }

/* 		Tabs
*/

    if (ch == '\t') {
        HTTabStop * tab;
	int target;	/* Where to tab to */
	int here = line->size + line->offset +indent;
        if (style->tabs) {	/* Use tab table */
	    for (tab = style->tabs;
	    	tab->position <= here;
		tab++)
		if (!tab->position) {
		    new_line(text);
		    return;
		}
	    target = tab->position;
	} else if (text->in_line_1) {	/* Use 2nd indent */
	    if (here >= style->leftIndent) {
	        new_line(text); /* wrap */
		return;
	    } else {
	        target = style->leftIndent;
	    }
	} else {		/* Default tabs align with left indent mod 8 */
#ifdef DEFAULT_TABS_8
	    target = ((line->offset + line->size + 8) & (-8))
	    		+ style->leftIndent;
#else
	    new_line(text);
	    return;
#endif
	}
	if (target > (COLS-1) - style->rightIndent) {
	    new_line(text);
	    return;
	} else {
            text->permissible_split = line->size;	/* Can split here */
	    if (line->size == 0) line->offset = line->offset + target - here;
	    else for(; here<target; here++) {
                line->data[line->size++] = ' ';	/* Put character into line */
	    }
	    return;
	}
	/*NOTREACHED*/
    } /* if tab */ 

    
    if (ch==' ') {
        text->permissible_split = line->size;	/* Can split here */
    }

/*	Check for end of line
*/    
    if (indent + line->offset + line->size + style->rightIndent
    		>= (COLS-1)) {
        if (style->wordWrap) {
	    split_line(text, text->permissible_split);
	    if (ch==' ') return;	/* Ignore space causing split */
	} else new_line(text);
    }

/*	Insert normal characters
*/
    if (ch == HT_NON_BREAK_SPACE) {
        ch = ' ';
    }
    {
        HTLine * line = text->last_line;	/* May have changed */
        HTFont font = style->font;
        line->data[line->size++] =	/* Put character into line */
           font & HT_CAPITALS ? TOUPPER(ch) : ch;
        if (font & HT_DOUBLE)		/* Do again if doubled */
            HText_appendCharacter(text, HT_NON_BREAK_SPACE);
	    /* NOT a permissible split */ 
    }
}

/*		Anchor handling
**		---------------
*/
/*	Start an anchor field
*/
PUBLIC void HText_beginAnchor ARGS2(HText *,text, HTChildAnchor *,anc)
{
    char marker[100];

    TextAnchor * a = (TextAnchor *) malloc(sizeof(*a));
    
    if (a == NULL) outofmem(__FILE__, "HText_beginAnchor");
    a->hightext = NULL;
    a->start = text->chars + text->last_line->size;

    a->line_pos = text->last_line->size;
    if (text->last_anchor) {
        text->last_anchor->next = a;
    } else {
        text->first_anchor = a;
    }
    a->next = 0;
    a->anchor = anc;
    a->extent = 0;
    text->last_anchor = a;
    
    if (HTAnchor_followMainLink((HTAnchor*)anc)) {
        a->number = ++(text->last_anchor_number);
    } else {
        a->number = 0;
    }

    /* if we are doing link_numbering add the link number */
    if(keypad_mode == LINKS_ARE_NUMBERED) {
	sprintf(marker,"[%d]", a->number);
        HText_appendText(text, marker);
	a->numbered = YES;
    } else {
	a->numbered = NO;
    }
}


PUBLIC void HText_endAnchor ARGS1(HText *,text)
{
    TextAnchor * a = text->last_anchor;
    char marker[100];
    if (a->number) {	 /* If it goes somewhere */
            a->extent += text->chars + text->last_line->size - a->start;
	    a->show_anchor = YES;
    } else {
	    a->show_anchor = NO;
	    a->extent = 0;
    }
 
}


PUBLIC void HText_appendText ARGS2(HText *,text, CONST char *,str)
{
    CONST char * p;
    for(p=str; *p; p++) {
        HText_appendCharacter(text, *p);
    }
}


PUBLIC void HText_endAppend ARGS1(HText *,text)
{
    int cur_line,cur_char;
    TextAnchor *anchor_ptr;
    HTLine *line_ptr;
    char tmp_string[512];
    char *cp;

    if(!text)
	return;

    new_line(text);
    
#ifdef DISPLAY_ON_FLY
    if (text->display_on_the_fly) {		/* Not finished? */
	text->display_on_the_fly = 0;
	text->next_line = text->last_line;	/* Bug fix after EvA 920117 */
	text->stale = NO;
    }
#endif /* DISPLAY_ON_FLY */

    /* get the hightext from the text by finding the char
     * position, then
     * bring the anchors in line with the text by adding the
     * text offset to each of the anchors
     */
     /* get first line */
     line_ptr = text->last_line->next;
     cur_char = line_ptr->size;;
     cur_line = 0;

     if(TRACE)
	fprintf(stderr,"Gridtext: Entering HText_endAppend\n");

     for(anchor_ptr=text->first_anchor; anchor_ptr; 
				             anchor_ptr=anchor_ptr->next) {

	  /* find the right line */
	 for(;anchor_ptr->start > cur_char; line_ptr = line_ptr->next, 
				    cur_char += line_ptr->size+1, cur_line++) 
		; /* null body */

	 anchor_ptr->line_pos = anchor_ptr->start-(cur_char - line_ptr->size);

	 /* strip off a spaces at the beginning if they exist */
         while(isspace(line_ptr->data[anchor_ptr->line_pos])) {
		anchor_ptr->line_pos++;
		anchor_ptr->extent--;
	 }

	 /* copy the link name into the data structure */
	 if(line_ptr->data && strlen(line_ptr->data) > anchor_ptr->line_pos
	    && anchor_ptr->extent > 0 && anchor_ptr->line_pos >= 0)
	     strncpy(tmp_string, &line_ptr->data[anchor_ptr->line_pos],
						          anchor_ptr->extent);
	 tmp_string[anchor_ptr->extent] = '\0';
	 StrAllocCopy(anchor_ptr->hightext, tmp_string);

         anchor_ptr->line_pos += line_ptr->offset;  /* add the offset */
         anchor_ptr->line_num  = cur_line;

	 if(TRACE)
	   fprintf(stderr,"GridText: adding link on line %d in HText_endAppend\n", cur_line);

 	 if(anchor_ptr == text->last_anchor)
	     break;
     }

}



/* 	Dump diagnostics to stderr
*/
PUBLIC void HText_dump ARGS1(HText *,text)
{
    fprintf(stderr, "HText: Dump called\n");
}
	

/*	Return the anchor associated with this node
*/
PUBLIC HTParentAnchor * HText_nodeAnchor ARGS1(HText *,text)
{
    return text->node_anchor;
}

/*				GridText specials
**				=================
*/
/*	Return the anchor with index N
**
**	The index corresponds to the number we print in the anchor.
*/
PUBLIC HTChildAnchor * HText_childNumber ARGS1(int,number)
{
    TextAnchor * a;
    for (a = HTMainText->first_anchor; a; a = a->next) {
        if (a->number == number) return(a->anchor);
    }
    return (HTChildAnchor *)0;	/* Fail */
}

/* HTGetLinkInfo returns some link info 
 */
PUBLIC int HTGetLinkInfo ARGS4(int, number, char *, hightext, char *, lname,
				HTAnchor **, www_destination)
{
    TextAnchor * a;
    char * cp;

    for (a = HTMainText->first_anchor; a; a = a->next) {
        if (a->number == number) {
	    if(a->numbered) {
	  	cp = strchr(a->hightext, ']');
	        strcpy(hightext, cp);
	    } else {
	        strcpy(hightext, a->hightext);
	    }
            *www_destination = HTAnchor_followMainLink(
                                               (HTAnchor *)a->anchor);
            strcpy(lname, HTAnchor_address(*www_destination));
	    return(YES);
	}
    }
    return(NO);
}

/* HText_getNumOfLines returns the number of lines in the
 * current document
 */
PUBLIC int HText_getNumOfLines()
{
     return(HTMainText->lines);
}

/* HText_getTitle returns the title of the
 * current document
 */
PUBLIC char * HText_getTitle()
{
    if(HTAnchor_title(HTMainText->node_anchor))
        /* return(HTMainText->title); */
	return(HTAnchor_title(HTMainText->node_anchor));
    else
	return("");
}

/* HText_pageDisplay displays a screen of text
 * starting from the line 'line_num'-1
 * this is the primary call for lynx
 */
extern char is_www_index;

PUBLIC void HText_pageDisplay ARGS2(int,line_num, char *, target)
{
    display_page(HTMainText, line_num-1, target);

    is_www_index = HTAnchor_isIndex(HTMainAnchor);
} 

PUBLIC void HText_setStale ARGS1(HText *,text)
{
    text->stale = YES;
}

PUBLIC void HText_refresh ARGS1(HText *,text)
{
    if (text->stale) display_page(text, text->top_of_screen, "");
}

PUBLIC int HText_sourceAnchors ARGS1(HText *,text)
{
    return text->last_anchor_number;
}

PUBLIC BOOL HText_canScrollUp ARGS1(HText *,text)
{
    return text->top_of_screen != 0;
}

PUBLIC BOOL HText_canScrollDown ()
{
    HText * text = HTMainText;

    char * title = text->title;
    return (text->top_of_screen + LINES-2 - (title? TITLE_LINES:0) ) <
    		text->lines;
}

/*		Scroll actions
*/
PUBLIC void HText_scrollTop ARGS1(HText *,text)
{
    display_page(text, 0, "");
}

PUBLIC void HText_scrollDown ARGS1(HText *,text)
{
    display_page(text, text->top_of_screen + LINES -1, "");
}

PUBLIC void HText_scrollUp ARGS1(HText *,text)
{
    display_page(text, text->top_of_screen - LINES +1, "");
}

PUBLIC void HText_scrollBottom ARGS1(HText *,text)
{
    display_page(text, text->lines - LINES +1, "");
}


/*		Browsing functions
**		==================
*/

/* Bring to front and highlight it
*/

PRIVATE int line_for_char ARGS2(HText *,text, int,char_num)
{
    int line_number =0;
    int characters = 0;
    HTLine * line = text->last_line->next;
    for(;;) {
	if (line == text->last_line) return 0;	/* Invalid */
        characters = characters + line->size + 1;
	if (characters > char_num) return line_number;
	line_number ++;
	line = line->next;
    }
}

PUBLIC BOOL HText_select ARGS1(HText *,text)
{
    if (text != HTMainText) {
        HTMainText = text;
	HTMainAnchor = text->node_anchor;
	  /* let lynx do it */
	/* display_page(text, text->top_of_screen, ""); */
    }
    return YES;
}

PUBLIC BOOL HText_selectAnchor ARGS2(HText *,text, HTChildAnchor *,anchor)
{
    TextAnchor * a;

/* This is done later, hence HText_select is unused in GridText.c
   Should it be the contrary ? @@@
    if (text != HTMainText) {
        HText_select(text);
    }
*/

    for(a=text->first_anchor; a; a=a->next) {
        if (a->anchor == anchor) break;
    }
    if (!a) {
        if (TRACE) fprintf(stderr, "HText: No such anchor in this text!\n");
        return NO;
    }

    if (text != HTMainText) {		/* Comment out by ??? */
        HTMainText = text;		/* Put back in by tbl 921208 */
	HTMainAnchor = text->node_anchor;
    }

    {
	 int l = line_for_char(text, a->start);
	if (TRACE) fprintf(stderr,
	    "HText: Selecting anchor [%d] at character %d, line %d\n",
	    a->number, a->start, l);

	if ( !text->stale &&
	     (l >= text->top_of_screen) &&
	     ( l < text->top_of_screen + LINES-1))
	         return YES;

       /* display_page(text, l - (LINES/3), ""); */	/* Scroll to it */
	www_search_result = l - (LINES/3);  /* put in global variable */
    }
    
    return YES;
}
 

/*		Editing functions		- NOT IMPLEMENTED
**		=================
**
**	These are called from the application. There are many more functions
**	not included here from the orginal text object.
*/

/*	Style handling:
*/
/*	Apply this style to the selection
*/
PUBLIC void HText_applyStyle ARGS2(HText *, me, HTStyle *,style)
{
    
}


/*	Update all text with changed style.
*/
PUBLIC void HText_updateStyle ARGS2(HText *, me, HTStyle *,style)
{
    
}


/*	Return style of  selection
*/
PUBLIC HTStyle * HText_selectionStyle ARGS2(
	HText *,me,
	HTStyleSheet *,sheet)
{
    return 0;
}


/*	Paste in styled text
*/
PUBLIC void HText_replaceSel ARGS3(
	HText *,me,
	CONST char *,aString, 
	HTStyle *,aStyle)
{
}


/*	Apply this style to the selection and all similarly formatted text
**	(style recovery only)
*/
PUBLIC void HTextApplyToSimilar ARGS2(HText *,me, HTStyle *,style)
{
    
}

 
/*	Select the first unstyled run.
**	(style recovery only)
*/
PUBLIC void HTextSelectUnstyled ARGS2(HText *,me, HTStyleSheet *,sheet)
{
    
}


/*	Anchor handling:
*/
PUBLIC void		HText_unlinkSelection ARGS1(HText *,me)
{
    
}

PUBLIC HTAnchor *	HText_referenceSelected ARGS1(HText *,me)
{
     return 0;   
}


PUBLIC int HText_getTopOfScreen ()
{
      HText * text = HTMainText;
      return text->top_of_screen;
}

PUBLIC int HText_getLines ARGS1(HText *,text)
{
      return text->lines;
}

PUBLIC HTAnchor *	HText_linkSelTo ARGS2(HText *,me, HTAnchor *,anchor)
{
    return 0;
}

#ifdef NOT
PUBLIC int do_www_search ARGS1(char *,newfile)
{
       char searchstring[120];
       char *ss_ptr;  /* pointer to first non space character in searchstring */
       char *cp, *cp2;	

       *searchstring='\0';
       statusline("Enter a database search string: ");
       Gophergetstr(searchstring);

	/* get rid of opening spaces */
       ss_ptr = searchstring;
       while(isspace(*ss_ptr)) ss_ptr++;

       if(strlen(ss_ptr) == 0)
	   return(NO);

       for(cp = ss_ptr; *cp != '\0'; cp++) 
		/* replace groups of spaces with a single '+' */
	   if(*cp == ' ') {
	      *cp = '+';
	      while(isspace(*(++cp))) {
		  for(cp2=cp+1; *cp2 != '\0'; cp2++)
		      *(cp2-1) = *cp2;
		  *(cp2-1) = '\0';
	      }
	   }

	/* put up searching message */
       statusline(WWW_WAIT_MESSAGE);

       strcat(newfile, "?");
       strcat(newfile, ss_ptr);

       return(YES);

}
#endif NOT

PUBLIC int do_www_search ARGS1(char *,newfile)
{
       char searchstring[120];

       *searchstring='\0';
       statusline("Enter a database search string: ");
       if(Gophergetstr(searchstring) < 0 || *searchstring == '\0')
	  return(NO);

       statusline(WWW_WAIT_MESSAGE);

       if (HTSearch(searchstring, HTMainAnchor)) {
           strcpy(newfile, HTAnchor_address(HTMainAnchor)); 
	   printf("\ndo_www_search: newfile: %s\n",newfile);
	   fflush(stdout);
	   sleep(2);
	   return(YES);
	}


       return(NO);
}

PUBLIC void print_wwwfile_to_fd ARGS1(FILE *,fp)
{

    HTLine * line = HTMainText->last_line->next;

    for(;; line = line->next) {

	fprintf(fp,"%s%s\n",SPACES(line->offset),line->data);

	if(line == HTMainText->last_line)
	    break;
    }
}

PUBLIC int www_user_search ARGS2(int,start_line, char *,target)
{
    register HTLine * line = HTMainText->last_line->next;
    register int count;
    BOOLEAN Found=NO;
    extern BOOLEAN case_sensitive;

	/* advance to the start line */
    for(count=1; count <= start_line; line=line->next, count++)
	; /* null */

    for(;;) {
	if(case_sensitive && strstr(line->data, target)) {
	    Found = YES;
	    break;
	} else if(!case_sensitive && mystrstr(line->data, target)) {
	    Found = YES;
	    break;

	} else if(line == HTMainText->last_line) {  /* next line */
	    break;
	} else {			/* end */
	    line = line->next;
	    count++;
	}
    }

	/* search from the beginning */
    if(!Found) {
	line = HTMainText->last_line->next; /* set to first line */
	count = 1;

        for(;;) {
	    if(case_sensitive && strstr(line->data, target)) {
	        Found = YES;
	        break;
	    } else if(!case_sensitive && mystrstr(line->data, target)) {
	        Found = YES;
	        break;

	    } else if(count > start_line) {  /* next line */
	        break;
	    } else {			/* end */
	        line = line->next;
		count++;
	    }
        } 
    }

   if(Found)
	return(count);
   else
	return(-1);
}


PUBLIC  void  user_message ARGS2(char *,message, char *,argument) 
{
    char temp[256];

    sprintf(temp, message, argument);

    statusline(temp);
    sleep(2);
   
}

PUBLIC void HTLink_setMaker ARGS1(char *,href)
{
    /* make the href the owner_address */
    strcpy(owner_address, href);
}
